/*
 * Copyright 2021 Gao's lab, Peking University, CCME. All rights reserved.
 *
 * NOTICE TO LICENSEE:
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef __META_CUH__
#define __META_CUH__
#include "../common.cuh"
#include "../control.cuh"
#include "vector"
#include <iostream>
#include "Grid.cu"
#include <thrust/host_vector.h>
#include "SwitchFunction.cuh"
#include "../collective_variable/collective_variable.cuh"

using namespace std;
struct META 
{
	using Gdata = thrust::host_vector<float>; ///< Grid's data storage as field.
	Grid<Gdata>* grid = nullptr;
	Grid<float>* potential_grid = nullptr;
	using Axis = std::vector<float>; ///< Axis with upper/lower boundary.
	struct Hill
	{
		/**
		 * @brief  Constructs a multidimensional Hill (Gaussian)
		 * @param[in] center Hill center.
		 * @param[in] inv_w Hill width(inverted).
		 * @param[in] period periodic box length.
		 * @param[in] height Hill height.
		 */
		Hill(const Axis &centers, const Axis &inv_w, const Axis &period, const float &theight);
		/**
		 * @brief Evaluate the Gaussian hill
		 * The F_old will be changed
		 * @param[in] values CV values for Hill as axis.
		 */
		Gdata CalcHill(const Axis &values);
		vector<GaussianSF> gsf; ///< Hill center and width.
		float height;		///< Hill height.
		float potential;	///< Hill potential.
	};
	vector<Hill> hills;
	bool usegrid = true;//false; ///< True or false.
	void SetGrid(void);
	void AddHill(const string &fn, Hill &hill);
	void Estimate(const Axis& values,vector<Hill>& hills);
	int is_initialized = 0;

	//作用的CV
	//   COLLECTIVE_VARIABLE_PROTOTYPE *cv;
	CV_LIST cvs;
	// using Gdata = VectorXd;       ///< Internal data storage of Grid.
	int ndim = 1; ///< default 1D metadynamics
	// ndim-Meta的边界和细度记录
	float border_potential_height= 100000.;//cv边界近似无限高势垒
	vector<float> border_lower;
	vector<float> border_upper;
	vector<float> cv_period;
	float *cv_mins;
	float *cv_maxs;
	vector<float> cv_deltas;
	float *cv_periods;
	float *grid_potential;
	float current_potential;
	float *cv_sigmas;
	int *n_grids;
	int total_grids = 0;

	float **d_cv_values;
	VECTOR **d_crd_grads;
	VECTOR **d_box_grads;

	//多次运行SPONGE相关的记录存储操作
	char read_potential_file_name[256];
	char write_potential_file_name[256];
	void Read_Potential(CONTROLLER *controller);
	void Write_Potential();

	float *cutoff;		 //查表的范围（通常是3sigma）
	float height;		 //通过well-temperature修正的reweight过后的势场高度
	float height_0;		 //初始势场高度

	float welltemp_factor = 1000000000.; // Biasfactor无限大时，即为普通的Meta无时间衰减，因此预留一个大值。
	int is_welltemp = 0;
	float sys_temp = 0.; //系统温度

	void Meta_Force_With_Energy_And_Virial(int atom_numbers, VECTOR *frc, int need_potential, int need_pressure, float *d_potential, float *d_virial, VECTOR box_length);

	char module_name[CHAR_LENGTH_MAX];
	void Initial(CONTROLLER *controller, COLLECTIVE_VARIABLE_CONTROLLER *cv_controller, char *module_name = NULL);
	void AddPotential(float sys_temp, int steps);
	void getHeight(float sys_temp);
	void computeReweightingFactor(float sys_temp);
	void getReweightingBias(float temp);

	void Potential_and_derivative(const int need_potential);							    //?��?????Potential??DPotential
	void Border_derivative(float *upper, float *lower, float *cutoff, float *Dpotential_local); //????Upper_Wall??��??

	float potential_local = 0.;	   //存储float Potential()
	float *Dpotential_local = nullptr; //存储float* DPotential()
	int potential_update_interval;
	float rct = 0.;
	float rbias = 0.; // Rw_Potential：rbias=potential(x)-rct;
	float bias = 0.;  // bias;
	void Do_Metadynamics(int atom_numbers, UNSIGNED_INT_VECTOR *uint_crd, VECTOR scaler, VECTOR *crd, VECTOR box_length,
			     int step, int need_potential, int need_pressure, VECTOR *frc, float *potential, float *virial, float sys_temp);
};
#endif // META
