﻿#include "DiamondStructure.cuh"

void DIAMOND_STRUCTURE::Initial(int system_atom_numbers, VECTOR boxlength, CONTROLLER *controller, char *module_name)
{
	if (module_name == NULL)
	{
		strcpy(this->module_name, "diamond_structure");
	}
	else
	{
		strcpy(this->module_name, module_name);
	}
	printf("START INITIALIZING DIAMOND_STRUCTURE:\n");
//	if (controller[0].Command_Exist("cut"))
//		cutoff = atof(controller[0].Command("cut"));

	if (controller[0].Command_Exist(this->module_name, "atom_in_file"))
	{
		FILE *temp_in = NULL;
		Open_File_Safely(&temp_in, controller[0].Command(this->module_name, "atom_in_file"), "r");
		fscanf(temp_in, "%d", &atom_numbers);
		if (atom_numbers > system_atom_numbers)
		{
			printf("	Error:Diamond_structure atom numbers %d is bigger than system atom numbers %d! Please Check!\n", atom_numbers, system_atom_numbers);
			getchar();
		}
		Malloc_Safely((void**)&h_atom_serial, sizeof(int)*atom_numbers);
		Cuda_Malloc_Safely((void**)&d_atom_serial, sizeof(int)*atom_numbers);

		if (controller[0].Command_Exist(this->module_name, "max_neighbor_numbers"))
		{
			max_neighbor_numbers = atoi(controller[0].Command(this->module_name, "max_neighbor_numbers"));
		}
		if (controller[0].Command_Exist(this->module_name, "neighbor_list_cutoff"))
		{
			dis_nl = atof(controller[0].Command(this->module_name, "neighbor_list_cutoff"));
		}

		Malloc_Safely((void**)&h_f2p_serial, sizeof(int)*system_atom_numbers);
		Cuda_Malloc_Safely((void**)&d_f2p_serial, sizeof(int)*system_atom_numbers);
		for (int i = 0; i < system_atom_numbers; i = i + 1)
		{
			h_f2p_serial[i] = -1;
		}
		for (int i = 0; i < atom_numbers; i = i + 1)
		{
			fscanf(temp_in, "%d\n", &h_atom_serial[i]);
			h_f2p_serial[h_atom_serial[i]] = i;
		}

		cudaMemcpy(d_f2p_serial, h_f2p_serial, sizeof(int)*system_atom_numbers, cudaMemcpyHostToDevice);
		cudaMemcpy(d_atom_serial, h_atom_serial, sizeof(int)*atom_numbers, cudaMemcpyHostToDevice);

		Malloc_Safely((void**)&h_nl, sizeof(ATOM_GROUP)*atom_numbers);
		Cuda_Malloc_Safely((void**)&d_nl, sizeof(ATOM_GROUP)*atom_numbers);

		Malloc_Safely((void**)&h_nl_f, sizeof(ATOM_GROUP)*atom_numbers);
		Cuda_Malloc_Safely((void**)&d_atom_ucrd, sizeof(UNSIGNED_INT_VECTOR)*atom_numbers);
		Malloc_Safely((void**)&atom_crd, sizeof(UNSIGNED_INT_VECTOR)*atom_numbers);
		Malloc_Safely((void**)&fnl, sizeof(ATOM_GROUP)*atom_numbers);
		Malloc_Safely((void**)&snl, sizeof(Second_Neighbor_GROUP)*atom_numbers);
		Malloc_Safely((void**)&CNA, sizeof(CNA_GROUP)*atom_numbers);
		Malloc_Safely((void**)&diamond_structure_type, sizeof(DIAMOND_GROUP)*atom_numbers);	
		Malloc_Safely((void**)&temp_neighbor, sizeof(int)*max_neighbor_numbers);
		Malloc_Safely((void**)&CNA_temp, sizeof(int)* 11);
		for (int i = 0; i < atom_numbers; i = i + 1)
		{
			h_nl[i].atom_numbers = 0;
			Cuda_Malloc_Safely((void**)&h_nl[i].atom_serial, sizeof(int)*max_neighbor_numbers);
			h_nl_f[i].atom_numbers = 0;
			Malloc_Safely((void**)&h_nl_f[i].atom_serial, sizeof(int)*max_neighbor_numbers);
			fnl[i].atom_numbers = 0;
			Malloc_Safely((void**)&fnl[i].atom_serial, sizeof(int)*fnl_neighbornumber);
			Malloc_Safely((void**)&snl[i].atom_serial, sizeof(int)*snl_neighbornumber);
			Malloc_Safely((void**)&CNA[i].CNA_style, sizeof(int)*CNA_style_numbers);
			Malloc_Safely((void**)&diamond_structure_type[i].fnl_type, sizeof(int)*fnl_neighbornumber);
			Malloc_Safely((void**)&diamond_structure_type[i].snl_type, sizeof(int)*snl_neighbornumber);

		}
		cudaMemcpy(d_nl, h_nl, sizeof(ATOM_GROUP)*atom_numbers, cudaMemcpyHostToDevice);
		for (int i = 0; i < atom_numbers; i = i + 1)
		{
			Malloc_Safely((void**)&h_nl[i].atom_serial, sizeof(int)* max_neighbor_numbers);
		}

		Malloc_Safely((void**)&structure_counts, sizeof(int)* 7);
		for (int i = 0; i < 7; i = i + 1)
		{
			structure_counts[i] = 0;
		}
		if (controller[0].Command_Exist(this->module_name, "need_cluster_analysis"))
		{
			need_clusteranalysis = atoi(controller[0].Command(this->module_name, "need_cluster_analysis"));
		}
		if (need_clusteranalysis)
		{
			if (controller[0].Command_Exist(this->module_name, "cluster_cutoff"))
			{
				cluster_cutoff = atof(controller[0].Command(this->module_name, "cluster_cutoff"));
			}
			if (controller[0].Command_Exist(this->module_name, "onlyice"))
			{
				onlyice = atoi(controller[0].Command(this->module_name, "onlyice"));
			}
			max_ca_numbers = atom_numbers;//cluster_neighbor列表宽默认取粒子数
			if (controller[0].Command_Exist(this->module_name, "max_cluster_neighbor_numbers"))
			{
				max_ca_numbers = atoi(controller[0].Command(this->module_name, "max_cluster_neighbor_numbers"));
			}
			Malloc_Safely((void**)&h_ca_nl, sizeof(ATOM_GROUP)*atom_numbers);
			for (int i = 0; i < atom_numbers; i = i + 1)
			{
				h_ca_nl[i].atom_numbers = 0;
				Malloc_Safely((void**)&h_ca_nl[i].atom_serial, sizeof(int)*max_ca_numbers);
			}
			Cuda_Malloc_Safely((void**)&d_ca_nl, sizeof(ATOM_GROUP)*atom_numbers);
			//cudaMemcpy(d_ca_nl, h_ca_nl, sizeof(ATOM_GROUP)*atom_numbers, cudaMemcpyHostToDevice);这行代码抄的之前的版本，可执行，但是d_ca_nl调用有误，用下面的方式取代
			cudaMemcpy(d_ca_nl, d_nl, sizeof(ATOM_GROUP)*atom_numbers, cudaMemcpyDeviceToDevice);
			Malloc_Safely((void**)&temp_ca_neighbor, sizeof(int)*max_ca_numbers);
			Malloc_Safely((void**)&cluster_id, sizeof(int)*atom_numbers);//宽度为atom_number的列表，计算时置空为-1，按地图染色原理添加cluster_id
			Malloc_Safely((void**)&cluster_list, sizeof(int)*atom_numbers);//按照cluster_id排列的cluster所含元素数量列表
			Malloc_Safely((void**)&degenerate_id, sizeof(int)*atom_numbers);//需要进行简并操作的cluster_id列表
		}
		is_initialized = 1;
		controller[0].Step_Print_Initial("cubicity", "%.3f");
		controller[0].Step_Print_Initial("ice_ratio", "%.3f");

		printf("	Diamond_structure atom numbers is %d\n", atom_numbers);
		printf("	Max neighbor numbers is %d\n", max_neighbor_numbers);
		printf("	Neighbor List Cutoff is %.3f\n", dis_nl);
	}
	else
	{
		printf("	Warnning: No atom_in_file information loaded, atom_numbers will be set to 0!\n");
		need_diamondstructure = 0;
	}
	

	if (need_clusteranalysis)
	{
		printf("	ClusterAnalysis Mode is On\n");
		printf("	ClusterAnalysis Cutoff is %.3f\n", cluster_cutoff);
		if (onlyice)
		{
			printf("		Only Ice Molecules are considered.\n");
		}
		printf("	Cluster Neighbor Numbers are less than %d\n", max_ca_numbers);
		controller[0].Step_Print_Initial("ClusterMax", "%d");
		controller[0].Step_Print_Initial("NCluster", "%d");
	}
	printf("END INITIALIZING DIAMOND_STRUCTURE\n");
}

__global__ void Copy_NeighborList_From_FullList_To_PartialList_D(const int partial_atom_numbers, const int *partial_to_full_atom_serial, const float dis_nl,
	ATOM_GROUP *partial_list, const int *full_to_partial_atom_serial, const ATOM_GROUP *full_list, const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler, const float max_neighbor_numbers)
{
	int patom_i = blockDim.x*blockIdx.x + threadIdx.x;
	if (patom_i < partial_atom_numbers)
	{
		ATOM_GROUP nl = full_list[partial_to_full_atom_serial[patom_i]];
		ATOM_GROUP pnl = partial_list[patom_i];
		pnl.atom_numbers = 0;
		int f2p_serial = -1;
		// nl.atom_numbers 替代为 max_neighbor_numbers
		//for (int i = 0; i < max_neighbor_numbers; i = i + 1)
		for (int i = 0; i < nl.atom_numbers; i = i + 1)
		{
			f2p_serial = full_to_partial_atom_serial[nl.atom_serial[i]];

			if (f2p_serial != -1)
			{
				VECTOR dr = Get_Periodic_Displacement(system_ucrd[partial_to_full_atom_serial[patom_i]], system_ucrd[partial_to_full_atom_serial[f2p_serial]], scaler);
				float dis = sqrtf(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
				if (dis < dis_nl)
				{				
					pnl.atom_serial[pnl.atom_numbers] = f2p_serial;
					pnl.atom_numbers = pnl.atom_numbers + 1;
				}
			}
		}
		partial_list[patom_i] = pnl;
	}
}

void DIAMOND_STRUCTURE::Refresh_Partial_Neighbor_List(const ATOM_GROUP *full_nl, const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler)
{
	Copy_NeighborList_From_FullList_To_PartialList_D << <ceilf((float)atom_numbers / 32), 32 >> >
		(atom_numbers, d_atom_serial, dis_nl, d_nl, d_f2p_serial, full_nl, system_ucrd, scaler, max_neighbor_numbers);
	cudaMemcpy(h_atom_serial, d_atom_serial, sizeof(int)*atom_numbers, cudaMemcpyDeviceToHost);
	int neighbor_numbers = 0;
	*temp_neighbor = NULL;
	int *h_neighbor = NULL;
	
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		cudaMemcpy(&neighbor_numbers, &d_nl[i].atom_numbers, sizeof(int), cudaMemcpyDeviceToHost);
		//if (neighbor_numbers>0)
		//{
		//	printf("\n NL:%d\n", neighbor_numbers);
		//}
		//保证此处数组不溢出
		/*
		if (neighbor_numbers>max_neighbor_numbers)
		{
			printf(" atom serial %d has %d neighbors, please use a larger max_neighbor_numbers instead of %d\n", i, neighbor_numbers, max_neighbor_numbers);
			getchar();
		}
		*/
		cudaMemcpy(&h_neighbor, &d_nl[i].atom_serial, sizeof(int*), cudaMemcpyDeviceToHost);
		//printf("   %d %d\n", d_neighbor[0], d_neighbor[1]);
		cudaMemcpy(temp_neighbor, h_neighbor, sizeof(int)*neighbor_numbers, cudaMemcpyHostToHost);
		//cudaMemcpy(&h_nl_f[i].atom_serial, &d_nl[i].atom_serial, sizeof(int)*neighbor_numbers, cudaMemcpyDeviceToHost);
		for (int j = 0; j < max_neighbor_numbers; j = j + 1)
		{
			//	printf("   %d\n", h_atom_serial[neighbor[j]]);
			//printf("   %d %d %d\n", j, neighbor_numbers, temp_neighbor[j]);
			if (j < neighbor_numbers)
			{
				h_nl_f[i].atom_serial[j] = temp_neighbor[j];
				//h_nl_f[i].atom_serial[j] = 1;
			}
			else
			{
				h_nl_f[i].atom_serial[j] = -1;//指针为-1不指向原子，用于占位
			}
		}
		h_nl_f[i].atom_numbers = neighbor_numbers;
		//检查近邻半表是否复制正确
		/*
		printf("\n atom %d, serial %d ,neighbor numbers %d:\n", i, h_atom_serial[i], h_nl_f[i].atom_numbers);
		for (int j = 0; j < neighbor_numbers; j = j + 1)
		{
			printf("   %d", h_nl_f[i].atom_serial[j]);
		}
		*/
	}

//将对角表还原为全表
	for (int i = 0; i < atom_numbers; i = i + 1)
		{
			int int_i = i;
			int atom_i_neighbor_numbers = h_nl_f[int_i].atom_numbers;
			for (int j = 0; j < atom_i_neighbor_numbers; j = j + 1)
			{
				int int_j = h_nl_f[int_i].atom_serial[j];
				if (int_j>int_i)
				{
					int atom_j_neighbor_numbers = h_nl_f[int_j].atom_numbers;
					h_nl_f[int_j].atom_serial[atom_j_neighbor_numbers] = int_i;
					h_nl_f[int_j].atom_numbers = atom_j_neighbor_numbers + 1;
				}
			}
		}
//检查近邻全表是否复制正确
	/*
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		printf("\n atom %d, serial %d ,neighbor numbers %d:\n",int_i, h_atom_serial[i], h_nl_f[i].atom_numbers);
		int atom_i_neighbor_numbers = h_nl_f[int_i].atom_numbers;
		//printf("%d %d %d\n", i, int_i, atom_i_neighbor_numbers);
		for (int j = 0; j < atom_i_neighbor_numbers; j = j + 1)
		{
			//printf("   %d", h_nl_f[i].atom_serial[j]);
		}
	}
	*/
}


//存储Diamond_structure部分原子的坐标
__global__ void Copy_Crd_From_System_To_Diamond_Structure(const int atom_numbers, const int *d_atom_serial, const UNSIGNED_INT_VECTOR *system_ucrd, 
	UNSIGNED_INT_VECTOR *ucrd)
{
	for (int i = threadIdx.x; i < atom_numbers; i = i + blockDim.x)
	{
		ucrd[i] = system_ucrd[d_atom_serial[i]];
	}
}
void DIAMOND_STRUCTURE::First_NeighborList(const UNSIGNED_INT_VECTOR *system_ucrd,const VECTOR scaler)
{
	//存储Diamond_structure部分原子的坐标
	Copy_Crd_From_System_To_Diamond_Structure << <atom_numbers / 128 + 1, 128 >> >(atom_numbers, d_atom_serial, system_ucrd, d_atom_ucrd);
	//cudaDeviceSynchronize();
	//getchar();
	cudaMemcpy(atom_crd, d_atom_ucrd, sizeof(UNSIGNED_INT_VECTOR)*atom_numbers, cudaMemcpyDeviceToHost);
	cudaMemcpy(h_atom_serial, d_atom_serial, sizeof(int)*atom_numbers, cudaMemcpyDeviceToHost);
	//检查是否复制成功
	/*
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = h_atom_serial[i];
		printf("%d %u %u %u\n", int_i, atom_crd[i].uint_x, atom_crd[i].uint_y, atom_crd[i].uint_z);
	}
	getchar();
	*/
	//求一阶近邻表
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		int atom_a = -1;
		int atom_b = -1;
		int atom_c = -1;
		int atom_d = -1;
		float dis_a = dis_nl;
		float dis_b = dis_nl;
		float dis_c = dis_nl;
		float dis_d = dis_nl;
		int atom_i_neighbor_numbers = h_nl_f[int_i].atom_numbers;
		//UNSIGNED_INT_VECTOR uint_crd_i = atom_crd[int_i];
		//printf("%d %d %d\n", i, int_i, atom_i_neighbor_numbers);
		//printf("%e %e %e\n", uint_crd_i.uint_x, uint_crd_i.uint_y, uint_crd_i.uint_z);
		for (int j = 0; j < atom_i_neighbor_numbers; j = j + 1)
		{
			int int_j = h_nl_f[int_i].atom_serial[j];
			//printf("%e %e %e\n", scaler.x, scaler.y, scaler.z);
			//printf("%d %u %u %u\n",int_i, atom_crd[int_i].uint_x, atom_crd[int_i].uint_y, atom_crd[int_i].uint_z);
			//printf("%d %u %u %u\n", int_j, atom_crd[int_j].uint_x, atom_crd[int_j].uint_y, atom_crd[int_j].uint_z);			
			VECTOR dr = Get_Periodic_Displacement(atom_crd[int_i], atom_crd[int_j], scaler);
			//printf("%f %f %f\n", dr.x, dr.y, dr.z);
			float dis = sqrtf(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
			float dr_temp = dis;
			int j_temp = int_j;
			float temp;
			int atom_temp;
			if (dr_temp < dis_a)
			{
				temp = dis_a;
				dis_a = dr_temp;
				dr_temp = temp;
				atom_temp = atom_a;
				atom_a = j_temp;
				j_temp = atom_temp;
			}
			if (dr_temp < dis_b)
			{
				temp = dis_b;
				dis_b = dr_temp;
				dr_temp = temp;
				atom_temp = atom_b;
				atom_b = j_temp;
				j_temp = atom_temp;
			}
			if (dr_temp < dis_c)
			{
				temp = dis_c;
				dis_c = dr_temp;
				dr_temp = temp;
				atom_temp = atom_c;
				atom_c = j_temp;
				j_temp = atom_temp;
			}
			if (dr_temp < dis_d)
			{
				temp = dis_d;
				dis_d = dr_temp;
				dr_temp = temp;
				atom_temp = atom_d;
				atom_d = j_temp;
				j_temp = atom_temp;
			}
			//printf("%d %d %f\n", j, int_j, dr_temp);
		}
		/*std::sort(dr2, dr2+atom_i_neighbor_numbers);
		int fnl_neighbornumber = 4;
		Malloc_Safely((void**)&fnl[i].atom_serial, sizeof(int)*fnl_neighbornumber);
		for (int j; j < fnl_neighbornumber; j = j + 1)
		{
			fnl[i].atom_serial[j] = dr2[j];
		}
		printf("%d %d %d %d %d\n", int_i, fnl[i].atom_serial[0], fnl[i].atom_serial[1], fnl[i].atom_serial[2], fnl[i].atom_serial[3]);
		getchar();*/
		fnl[i].atom_serial[0] = atom_a;
		fnl[i].atom_serial[1] = atom_b;
		fnl[i].atom_serial[2] = atom_c;
		fnl[i].atom_serial[3] = atom_d;
		//printf("%d %d %d %d %d %e %e %e %e\n", int_i, fnl[i].atom_serial[0], fnl[i].atom_serial[1], fnl[i].atom_serial[2], fnl[i].atom_serial[3], dis_a, dis_b, dis_c, dis_d);
		fnl[i].atom_numbers = fnl_neighbornumber;
	}
	//检查一阶表是否正确，正确的表只有前4位有数，第5位为空或负数
	/*
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		printf("\n atom %d, serial %d ,neighbor numbers %d:\n", int_i, h_atom_serial[i], fnl[i].atom_numbers);
		int atom_i_neighbor_numbers = fnl[i].atom_numbers;
		for (int j = 0; j < 5; j = j + 1)
		{
			printf("   %d\n", fnl[i].atom_serial[j]);
		}
	}
	*/
}
void DIAMOND_STRUCTURE::Second_NeighborList(const VECTOR scaler)
{
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		int snl_count = 0;
		float temp_sum = 0.;
		for (int j = 0; j < fnl_neighbornumber; j = j + 1)
		{
			int int_j = fnl[int_i].atom_serial[j];
			for (int k = 0; k < fnl_neighbornumber; k = k + 1)
			{
				//此处考虑一阶近邻不满位的情况，即int_j=-1，不执行计算赋值为-1，且此时temp_sum会超过12，不影响后续计算；
				int int_k = -1;
				if (int_j != -1)
				{
					int_k = fnl[int_j].atom_serial[k]; 
				}
				if (int_k != int_i)
				{
					snl[int_i].atom_serial[snl_count] = int_k;
					snl_count = snl_count + 1;
					VECTOR dr = Get_Periodic_Displacement(atom_crd[int_i], atom_crd[int_k], scaler);
					float dis = sqrtf(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
					temp_sum = temp_sum + dis;
				}
			}
		}
		snl[int_i].distance_average = temp_sum / snl_count;
		//检查二阶表是否正确，正确的表只有前12位有数，第13位为空或负数
		/*
		printf("\n  %d %d %d  %f %f\n",i, h_atom_serial[i], snl_count, temp_sum, snl[int_i].distance_average);
		for (int j = 0; j < 13; j = j + 1)
		{
			printf("   %d", snl[int_i].atom_serial[j]);
		}
		*/
	}
}
void DIAMOND_STRUCTURE::CNA_Analysis(const VECTOR scaler)
{
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		CNA[int_i].FCC_count = 0;
		CNA[int_i].HCP_count = 0;
		CNA[int_i].diamond_type = 0;
		float cutoff = snl[int_i].distance_average*(1+sqrtf(2))/2;
		for (int j = 0; j < snl_neighbornumber; j = j + 1)
		{
			//临时存储int_i与int_j构型判定的数列，11位，第1-3位为数量信息（421/422），第4-7位为共近邻Serial，第8-9位为共近邻第一对成键Serial，第10-11位位共近邻第二对成键Serial
			for (int k = 0; k < 11; k = k + 1)
			{			
				CNA_temp[k] = 0;
			}
			int int_j = snl[int_i].atom_serial[j];
			int common_neighbor_count = 0;
			int common_neighbor_bond_count = 0;
			int common_neighbor_chain_length = 1;
			for (int k = 0; k < snl_neighbornumber; k = k + 1)
			{
				int int_k = snl[int_i].atom_serial[k];
				if (int_k != int_j && int_k != int_i)
				{
					VECTOR dr1 = Get_Periodic_Displacement(atom_crd[int_i], atom_crd[int_k], scaler);
					float dis_ik = sqrtf(dr1.x*dr1.x + dr1.y*dr1.y + dr1.z*dr1.z);
					VECTOR dr2 = Get_Periodic_Displacement(atom_crd[int_j], atom_crd[int_k], scaler);
					float dis_jk = sqrtf(dr2.x*dr2.x + dr2.y*dr2.y + dr2.z*dr2.z);
					if (dis_ik < cutoff && dis_jk < cutoff)
					{
						common_neighbor_count = common_neighbor_count + 1;
						if (common_neighbor_count <= 4)
						{
							CNA_temp[common_neighbor_count + 2] = int_k;
						}
					}
				}
			}
			CNA_temp[0] = common_neighbor_count;
			if (common_neighbor_count == 4)
			{
				int bond_temp = 7;
				for (int m = 3; m < 6; m = m + 1)
				{
					int int_m = CNA_temp[m];
					for (int n = m + 1; n < 7; n = n + 1)
					{
						int int_n = CNA_temp[n];
						VECTOR dr = Get_Periodic_Displacement(atom_crd[int_m], atom_crd[int_n], scaler);
						float dis = sqrtf(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
						if (dis < cutoff)
						{
							common_neighbor_bond_count = common_neighbor_bond_count + 1;
							if (common_neighbor_bond_count <= 2)
							{
								CNA_temp[bond_temp] = int_m;
								bond_temp = bond_temp + 1;
								CNA_temp[bond_temp] = int_n;
								bond_temp = bond_temp + 1;
							}

						}
					}
				}
			}
			CNA_temp[1] = common_neighbor_bond_count;
			if (common_neighbor_bond_count == 2)
			{
				for (int m = 7; m < 11; m = m + 1)
				{
					int int_m = CNA_temp[m];
					for (int n = m + 1; n < 11; n = n + 1)
					{
						int int_n = CNA_temp[n];
						if (int_n == int_m && int_n >= 0)
						{
							common_neighbor_chain_length = common_neighbor_chain_length + 1;
						}

					}
				}
			}
			CNA_temp[2] = common_neighbor_chain_length;
			//检查临时存储表是否正确，第11位为空
			/*
			printf("\n   %d %d", int_i,int_j);
			for (int k = 0; k < 12; k = k + 1)
			{
				printf("   %d", CNA_temp[k]);
			}
			*/
			if (CNA_temp[0] == 4 && CNA_temp[1] == 2)
			{
				if (CNA_temp[2] == 1)
				{
					CNA[int_i].CNA_style[j] = 1;
					CNA[int_i].FCC_count = CNA[int_i].FCC_count + 1;
				}
				else if (CNA_temp[2] == 2)
				{
					CNA[int_i].CNA_style[j] = 2;
					CNA[int_i].HCP_count = CNA[int_i].HCP_count + 1;
				}
			}
			else
			{
				CNA[int_i].CNA_style[j] = 0;
			}
			//printf("\n   %d %d %d", int_i, int_j, CNA[int_i].CNA_style[j]);
		}
		if (CNA[int_i].FCC_count == 12)
		{
			CNA[int_i].diamond_type = 1;
		}
		else if (CNA[int_i].FCC_count == 6 & CNA[int_i].HCP_count == 6)
		{
			CNA[int_i].diamond_type = 2;
		}
		//printf("\n   %d %d %d %d", int_i, CNA[int_i].FCC_count, CNA[int_i].HCP_count, CNA[int_i].diamond_type);
	}
}

void DIAMOND_STRUCTURE::Diamond_Structure_Identification()
{
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int int_i = i;
		int flag_fnl = 0;
		int flag_snl = 0;
		diamond_structure_type[int_i].local_type = CNA[int_i].diamond_type;
		for (int j = 0; j < fnl_neighbornumber; j = j + 1)
		{
			int int_j = fnl[int_i].atom_serial[j];
			diamond_structure_type[int_i].fnl_type[j] = 0;
			//此处防止指向空邻居，type溢出值域
			if (int_j >= 0)
			{
				diamond_structure_type[int_i].fnl_type[j] = CNA[int_j].diamond_type;
				if (flag_fnl == 0 && CNA[int_j].diamond_type != 0)
				{
					flag_fnl = CNA[int_j].diamond_type;
				}
			}
		}
		for (int j = 0; j < snl_neighbornumber; j = j + 1)
		{
			int int_j = snl[int_i].atom_serial[j];
			diamond_structure_type[int_i].snl_type[j] = 0;
			//此处防止指向空邻居，type溢出值域
			if (int_j >= 0)
			{
				diamond_structure_type[int_i].snl_type[j] = CNA[int_j].diamond_type;
				if (flag_snl == 0 && CNA[int_j].diamond_type != 0)
				{
					flag_snl = CNA[int_j].diamond_type;
				}
			}
		}
		diamond_structure_type[int_i].readout_type = 0;
		if (diamond_structure_type[int_i].local_type == 1 || diamond_structure_type[int_i].local_type == 2)
		{
			diamond_structure_type[int_i].readout_type = 9 - 6 * diamond_structure_type[int_i].local_type;
		}
		else if (flag_fnl != 0)
		{
			diamond_structure_type[int_i].readout_type = 6 - 4 * flag_fnl;
		}
		else if (flag_snl != 0)
		{
			diamond_structure_type[int_i].readout_type = 3 - 2 * flag_snl;
		}
		//printf("\n %d %d", int_i, diamond_structure_type[int_i].readout_type);
		//异常流报错退出，预留
		if (diamond_structure_type[int_i].readout_type > 3 || diamond_structure_type[int_i].readout_type < -3)
		{
			for (int j = 0; j < 4; j = j + 1)
			{
				printf("  %d  ", diamond_structure_type[int_i].fnl_type[j]);
			}
			for (int j = 0; j < 12; j = j + 1)
			{
				printf("  %d  ", diamond_structure_type[int_i].snl_type[j]);
			}
			getchar();
		}
	}
}
void DIAMOND_STRUCTURE::Crystal_Statistics()
{
	for (int i = 0; i < 7; i = i + 1)
	{
		structure_counts[i] = 0;
	}
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		int type = diamond_structure_type[i].readout_type;
		structure_counts[3-type] = structure_counts[3-type] + 1;
	}
	//printf("\n Atom: %d Ic to Ih:  %d %d %d %d %d %d %d ", atom_numbers, structure_counts[0], structure_counts[1], structure_counts[2], structure_counts[3], structure_counts[4], structure_counts[5], structure_counts[6]);
	ice_ratio = (float)1. - (float)structure_counts[3] / atom_numbers;
	ice_count = atom_numbers - structure_counts[3];
	if (ice_ratio == 0.)
	{
		cubicity = 0.5;
	}
	else
	{
		cubicity = (float)(structure_counts[0] + structure_counts[1] + structure_counts[2]) / (atom_numbers - structure_counts[3]);
	}
	//printf("  cubicity :%.3f, ice_ratio :%.3f, ice_count :%d\n", cubicity, ice_ratio,ice_count);
}

void DIAMOND_STRUCTURE::Cluster_Cutoff_Neighbor_List(const ATOM_GROUP *full_nl, const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler)
{
	Copy_NeighborList_From_FullList_To_PartialList_D << <ceilf((float)atom_numbers / 32), 32 >> >
		(atom_numbers, d_atom_serial, cluster_cutoff, d_ca_nl, d_f2p_serial, full_nl, system_ucrd, scaler, max_ca_numbers);
	cudaMemcpy(h_atom_serial, d_atom_serial, sizeof(int)*atom_numbers, cudaMemcpyDeviceToHost);

	int ca_neighbor = 0;
	*temp_ca_neighbor = NULL;
	int *h_ca_neighbor = NULL;
	int ca_nl_numbers = 0;

	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		cudaMemcpy(&ca_neighbor, &d_ca_nl[i].atom_numbers, sizeof(int), cudaMemcpyDeviceToHost);
		if (ca_neighbor>max_ca_numbers)
		{
			printf(" atom serial %d has %d neighbors, please use a larger max_cluster_neighbor_numbers instead of %d\n", i, ca_neighbor, max_ca_numbers);
			getchar();
		}
		//非水判定，cluster判定不排除水或者原位判定为冰则执行操作
		ca_nl_numbers = 0;
		if (onlyice == 0 || diamond_structure_type[i].readout_type != 0)
		{
			cudaMemcpy(&h_ca_neighbor, &d_ca_nl[i].atom_serial, sizeof(int*), cudaMemcpyDeviceToHost);
			cudaMemcpy(temp_ca_neighbor, h_ca_neighbor, sizeof(int)*ca_neighbor, cudaMemcpyHostToHost);
			for (int j = 0; j < max_ca_numbers; j = j + 1)
			{
				//printf("   %d\n", h_atom_serial[neighbor[j]]);
				//printf("   %d %d %d\n", j, neighbor_numbers, temp_neighbor[j]);
				h_ca_nl[i].atom_serial[j] = -1;//指针为-1不指向原子，用于占位
				/*if (neighbor_numbers>0)
				{				
					printf("\n %d\n", neighbor_numbers);
				}*/

				if (j < ca_neighbor)
				{
					int int_j = temp_ca_neighbor[j];
					//printf("\n %d: %d\n", int_j, diamond_structure_type[int_j].readout_type);
					if (onlyice == 0 || diamond_structure_type[int_j].readout_type != 0)
					{
						h_ca_nl[i].atom_serial[ca_nl_numbers] = temp_ca_neighbor[j];
						ca_nl_numbers = ca_nl_numbers + 1;
						//printf("\n %d: %d\n", int_j, diamond_structure_type[int_j].readout_type);
						//h_nl_f[i].atom_serial[j] = 1;
					}
				}
			}
		}
		else
		{
			h_ca_nl[i].atom_serial[0] = -1;//排水判定时，水原子只有第一个元素为-1，其他为空
		}
		h_ca_nl[i].atom_numbers = ca_nl_numbers;
		//检查近邻半表是否复制正确
		/*
		if (need_clusteranalysis == 0 || diamond_structure_type[i].readout_type != 0)
		{
			printf("\n atom %d, serial %d ,type %d, neighbor numbers %d:\n", i, h_atom_serial[i], diamond_structure_type[i].readout_type, h_ca_nl[i].atom_numbers);
			for (int j = 0; j < h_ca_nl[i].atom_numbers; j = j + 1)
			{
				//printf("   %d", h_ca_nl[i].atom_serial[j]);
			}
		}
		*/
	}

	//将对角表还原为全表
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		if (onlyice == 0 || diamond_structure_type[i].readout_type != 0)
		{
			int int_i = i;
			int atom_i_neighbor_numbers = h_ca_nl[int_i].atom_numbers;
			for (int j = 0; j < atom_i_neighbor_numbers; j = j + 1)
			{
				int int_j = h_ca_nl[int_i].atom_serial[j];
				if (int_j>int_i)
				{
					//printf("\n %d: %d\n", int_j, diamond_structure_type[int_j].readout_type);
					if (onlyice == 0 || diamond_structure_type[int_j].readout_type != 0)
					{
						int atom_j_neighbor_numbers = h_ca_nl[int_j].atom_numbers;
						h_ca_nl[int_j].atom_serial[atom_j_neighbor_numbers] = int_i;
						h_ca_nl[int_j].atom_numbers = atom_j_neighbor_numbers + 1;
					}
				}
			}
		}
	}
	//检查近邻全表是否复制正确
	/*
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		if (onlyice == 0 || diamond_structure_type[i].readout_type != 0)
		{
			int int_i = i;
			printf("\n atom %d, serial %d ,neighbor numbers %d:\n", int_i, h_atom_serial[i], h_nl_f[i].atom_numbers);
			int atom_i_neighbor_numbers = h_nl_f[int_i].atom_numbers;
			//printf("%d %d %d\n", i, int_i, atom_i_neighbor_numbers);
			for (int j = 0; j < atom_i_neighbor_numbers; j = j + 1)
			{
				printf("   %d", h_nl_f[i].atom_serial[j]);
			}
		}
	}
	*/
}
void DIAMOND_STRUCTURE::Cluster_List()
{
	cluster_counts = 0;
	int temp_cluster_num = 0;
	int degenerate=0;//简并标记，0表示不简并，1,2,3表示简并度

	//重置列表
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		cluster_id[i] = -1;
		cluster_list[i] = 0;
	}

	for (int i = 0; i < atom_numbers; i = i + 1)//每个cluster一行
	{
		if (onlyice == 0 || diamond_structure_type[i].readout_type != 0)
		{
			degenerate = 0;
			*degenerate_id = NULL;//被简并的cluster_id
			temp_cluster_num = 0;

			if (cluster_id[i] == -1)//初始化，判定是否新增cluster
			{
				cluster_counts = cluster_counts + 1;
				cluster_id[i] = cluster_counts;
				temp_cluster_num = temp_cluster_num + 1;
			}
			else
			{
				temp_cluster_num = cluster_list[cluster_id[i]];
			}

			int color_id = cluster_id[i];//此时开始染色的cluster编号
			int father_id = color_id;//撞车的编号

			//int temp_cluster_new = 0;//撞车原有的团簇数

			for (int j = 0; j < max_ca_numbers; j = j + 1)//添加连通元素
			{
				int int_j = h_ca_nl[i].atom_serial[j];
				if (int_j >= 0)
				{
					if (cluster_id[int_j] == -1)//无归属元素进cluster
					{
						cluster_id[int_j] = color_id;
						temp_cluster_num = temp_cluster_num + 1;
					}
					else
					{
						if (cluster_id[int_j] < atom_numbers)//cluster兼并，将对应cluster的id全部收编反刷
						{
							int id_append = 1;
							if (cluster_id[int_j] < father_id)
							{
								father_id = cluster_id[int_j];
							}
							for (int k = 0; k < degenerate; k = k + 1)
							{
								if (cluster_id[int_j] == degenerate_id[k])
								{
									id_append = 0;
								}
							}
							if (id_append == 1)
							{
								degenerate_id[degenerate] = cluster_id[int_j];
								degenerate = degenerate + 1;//增加简并标记
							}
						}
					}
				}
			}
			//策反，刷墙			
			if (degenerate > 0)
			{
				//需要简并时，计数归0
				temp_cluster_num = 0;
				/*
				for (int k = 0; k < degenerate; k = k + 1)
				{
					if (color_id == degenerate_id[k])
					{
						printf("   %d exist\n", i);
					}
				}

				//简并度大于0时，至少father_id和color_id需要合并，将color_id追加到简并度中，新版本中已包括在degenerate_id中
				for (int j = 0; j < atom_numbers; j = j + 1)
				{
					if (cluster_id[j] == color_id)
					{
						cluster_id[j] = father_id;
						temp_cluster_num = temp_cluster_num + 1;
					}
					cluster_list[color_id] = 0;//置空cluster信息
				}
				*/
				for (int k = 0; k < degenerate; k = k + 1)
				{
					for (int j = 0; j < atom_numbers; j = j + 1)
					{
						if (cluster_id[j] == degenerate_id[k])//将相连通的cluster全部归入Id最小的cluster
						{
							cluster_id[j] = father_id;
							temp_cluster_num = temp_cluster_num + 1;
						}
					}
					cluster_list[degenerate_id[k]] = 0;//置空后拓cluster信息
				}
				cluster_list[father_id] = temp_cluster_num;//更新统计
				color_id = father_id;
			}
			else
			{
				cluster_list[color_id] = temp_cluster_num;
			}
		}
	}
	cluster_counts = 0;//去掉被简并的cluster
	largest_cluster_number = 0;
	largest_cluster_id = -1;
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		if (cluster_list[i]>0)
		{
			cluster_counts = cluster_counts + 1;
			//printf("   %d", cluster_list[i]);
			//记录最大的cluster
			if (cluster_list[i] > largest_cluster_number)
			{
				largest_cluster_number = cluster_list[i];
			}
		}
	}
	//printf("   \n");
	/*	
	//检查列表元素

	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		if (cluster_id[i] >= 0)
		{
			//printf("   %d: cluster %d\n", i, cluster_id[i]);
		}
	}

	
	//检查输出
	printf("   total:%d,	%d: %d", cluster_counts, largest_cluster_id, largest_cluster_number);
	for (int i = 0; i < cluster_counts; i = i + 1)
	{
		//printf("   %d\n", cluster_list[i]);
	}
	cluster_counts = 0;
	for (int i = 0; i < atom_numbers; i = i + 1)
	{
		if (cluster_list[i]>0)
		{
			cluster_counts = cluster_counts + 1;
			printf("   %d", cluster_list[i]);
		}
	}
	*/
	//打印cluster_list




	//getchar();
}