﻿/*
* Copyright 2021 Gao's lab, Peking University, CCME. All rights reserved.
*
* NOTICE TO LICENSEE:
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef CRYSTAL_D
#define CRYSTAL_D

#include "../common.cuh"
#include "../control.cuh"
#include<algorithm>
using namespace std;

//数据结构仿抄XRD3D，实现原理仿抄Ovito
//20210531，李茂东
//目前只写成CV，不能直接接入Meta

//二阶近邻表及截断Cutoff
struct Second_Neighbor_GROUP
{
	float distance_average;
	int *atom_serial;
};

//每对原子的CNA判定，可用于其他晶体分析
struct CNA_GROUP
{
	int FCC_count;
	int HCP_count;
	int *CNA_style;//每对粒子的判定，0为None,1为FCC，2为HCP
	int diamond_type;//每个粒子的判定，0为None,1为Ic，2为Ih
};

//原子及其一阶、二阶邻居原位构型，以及最终构型判定输出
struct DIAMOND_GROUP
{
	int readout_type;//输出层，diamond_type对应原子构型Ic=3/Ih=-3，一阶为Ic=2/Ih=-2，二阶为Ic=1/Ih=-1，水为0
	int local_type;//每个粒子的原位判定，0为None,1为Ic，2为Ih
	int *fnl_type;//一阶邻居的原位判定
	int *snl_type;//二阶邻居的原位判定
};

struct DIAMOND_STRUCTURE
{
public:

	int is_initialized = 0;//是否初始化
	//初始化参数
	int atom_numbers = 0;//注意，由于Diamond_structure计算很多时候只需要部分原子参与，
	//因此atom_numbers不等于体系真正的atom_numbers
	//UNSIGNED_INT_VECTOR *atom_uint_crd = NULL;//只针对Diamond_structure原子的坐标和力（Cv梯度）记录@CPU
	//UNSIGNED_INT_VECTOR *h_system_crd = NULL;//存储在CPU上的全原子坐标
	UNSIGNED_INT_VECTOR *d_atom_ucrd = NULL;//只针对Diamond_structure原子的坐标
	UNSIGNED_INT_VECTOR *atom_crd = NULL;//只针对Diamond_structure原子的坐标
	//VECTOR *atom_frc = NULL;//目前该CV没有反作用于Force
	int *h_atom_serial = NULL;//记录要计算Diamond_structure的原子序号
	int *d_atom_serial = NULL;
	int *h_f2p_serial = NULL;//体系所有原子向Diamond_structure原子的映射序号表，等于-1的体系原子不做Diamond_structure计算
	int *d_f2p_serial = NULL;
	float dis_nl = 4.5;//一阶近邻表搜索截断距离
	ATOM_GROUP *d_nl = NULL;//只记录要算Diamond_structure的原子的近邻表，序号也是按只有atom_numbers个原子进行索引，维度同NeighborList（对角半表）
	ATOM_GROUP *h_nl = NULL;//Diamond_structure的原子的近邻表的复原半表(这张表并不会更新，只用于初始化时生成d_nl)

	int *temp_neighbor = NULL;//临时存储邻居表
	ATOM_GROUP *h_nl_f = NULL;//Diamond_structure的原子的近邻表的复原全表，在此之后的操作都是CPU操作
	ATOM_GROUP *fnl = NULL;//Diamond_structure的一阶近邻表
	//float dis_fnl_max = 10.;//一阶近邻表搜索截断距离，已废弃，通过dis_nl实现
	Second_Neighbor_GROUP *snl = NULL;//Diamond_structure原子的二阶近邻表，记录12个次近邻表及平均距离；
	int *CNA_temp = NULL;//临时存储构型11位编码
	CNA_GROUP *CNA = NULL;//每对原子的CNA判定构型
	DIAMOND_GROUP *diamond_structure_type = NULL;//Diamond_structure原子，其中neighbor_type对应二阶近邻的构型FCC=1/HCP=-1，diamond_type对应原子构型Ic=3/Ih=-3


	//为了加速计算Diamond_structure所准备的近邻表和相关记录
	int need_diamondstructure = 1;//记录是否需要计算diamond_structure模块
	int max_neighbor_numbers = 50;//预留邻接表的宽度，在考虑多阶近邻时才需要扩增

	int fnl_neighbornumber = 4;//一阶邻居表4个位置
	int snl_neighbornumber = 12;//二阶邻居表12个位置
	//int CNA_serial_numbers = 9;//一阶邻居表4个位置
	int CNA_style_numbers = 12;//CNA判定型表，维度同snl

	//初始化Diamond_structure系统（Diamond_structure信息）
	char module_name[CHAR_LENGTH_MAX];
	//Diamond_structure模块初始化
	void Initial(int system_atom_numbers, VECTOR boxlength, CONTROLLER *controller, char *module_name = NULL);
	
	//重置Diamond_structure部分近邻表，将全表复制为小表
	void Refresh_Partial_Neighbor_List(const ATOM_GROUP *full_nl, const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler);

	//寻找一阶近邻；
	void First_NeighborList(const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler);
	//建二阶近邻表
	void Second_NeighborList(const VECTOR scaler);
	//计算CNA，该模块后续可包装为Crystal公用模块,此处的CNA不考虑后续存储打印每个int_i*int_j详细信息，维度为int_i*12;
	void CNA_Analysis(const VECTOR scaler);
	//包装层，存储每个粒子的原位判定，一阶、二阶邻居判定，及输出值，并汇总结果;
	void Diamond_Structure_Identification();
	////后处理之后的Diamond_structure，统计每种晶型的粒子数，可作为CV输出，计算Cubicity和Ice%
	void Crystal_Statistics();

	//输出CV，每个粒子的构型为diamond_structure_type[i].readout_type[3,2,1,0,-1,-2,-3]，以下是个数统计
	int *structure_counts = NULL;//最终的构型列表，7位分别是Ic\1stIc\2ndIc\Water\2ndIh\1stIh\Ih
	float cubicity = 0.000;//输出的CV，Ic_count/(Ic_count+Ih_count+eps)
	float ice_ratio = 0.000;//输出的CV，(Ic_count+Ih_count+eps)/atom_numbers
	int ice_count = 0;//非液相的粒子数,0923新增，可用于其他函数

	//2021.09.23新增ClusterAnalysis子模块，去除冰型判定标记可移植到广义的cluster定义
	//主原子为冰-邻接表-去掉邻接水
	//双重for循环构建cluster_neighbour_list【修改版，按地图染色原理，对cluster_id列表进行更新】
	//排序cluster_nl的元素个数，并输出
	int need_clusteranalysis = 0;//记录是否需要计算cluster模块，默认关闭
	float cluster_cutoff = 0;//cluster判定标准
	int onlyice = 0;//cluster判定只考虑冰
	int max_ca_numbers = 0;//预留cluster_analysis邻接表的宽度，默认取粒子数，可指定

	int *temp_ca_neighbor = NULL;//临时存储邻居表
	ATOM_GROUP *d_ca_nl = NULL;//只记录要算Diamond_structure的原子的近邻表，序号也是按只有atom_numbers个原子进行索引，维度同NeighborList（对角半表）
	ATOM_GROUP *h_ca_nl = NULL;//Diamond_structure的原子的近邻表的复原全表，在此之后的操作都是CPU操作
	//主程序NL半表转成全表，去掉非冰行&非冰邻居
	void Cluster_Cutoff_Neighbor_List(const ATOM_GROUP *full_nl, const UNSIGNED_INT_VECTOR *system_ucrd, const VECTOR scaler);

	int *cluster_id = NULL;//宽度为atom_number的列表，计算时置空为-1，按地图染色原理添加cluster_id
	void Cluster_List();
	int *cluster_list = NULL;//按照cluster_id排列的cluster所含元素数量列表
	int *degenerate_id = NULL; //被简并的cluster_id
	
	int largest_cluster_number=0;//最大cluster的粒子数
	int largest_cluster_id = 0;//最大cluster的索引，备调用
	int cluster_counts = 0;//构象中的cluster个数

};

#endif
